#!/usr/bin/perl
#V.Pohlers 5.1.15

#use Data::Dumper;
#$ARGV[0] ||= 'Gakken010.hex';

die <<help unless @ARGV;
Aufruf $0 datei.hex
GMC-4 Reassembler
help


my @hexcode;

#######################################
# Einlesen
#######################################

open IN, "<$ARGV[0]";
while (<IN>) {
	if (/^;/) { print; next } # Kommentarzeile
	chomp;
	s/ //g;	# Leerzeichen entfernen
	$_ = uc;
	push @hexcode, split '';
}
close IN;

#print @hexcode, "\n";	# Anzeige des Hex-Codes


#######################################
# Reassemblieren
#######################################


#for(my $i = 0; $i <= $#hexcode; $i++) {
#	print "$i: $hexcode[$i]\n";
#}

my $pc = 0; # program counter
my @codes = ();	# code: array of hash
my %marks = (); # die sprungadressen

#######################################
# Pass 1: Marken finden
#######################################

# alles reassemblieren
while ($pc < $#hexcode) {
	reass();
}

#print Dumper(\@code);
#print Dumper(\%marks);

#######################################
# Pass 2: Ausgabe
#######################################

for (my $i = 0; $i <= $#codes; $i++) {
#Ausgabe Assembler einfach
#	printf("%s: %-4s  %-10s ; %s\n", $codes[$i]{adr}, $codes[$i]{hex}, $codes[$i]{asm}, $codes[$i]{comment});
#Ausgabe in pseudo-c
	printf("%s: %-4s  %-7s %-34s ; %s\n", $codes[$i]{adr}, $codes[$i]{hex}, $syntaxmarks{$codes[$i]{adr}}, $codes[$i]{flagsyntax}.$codes[$i]{syntax}, $codes[$i]{comment});
#Ausgabe Assembler mit Marken
#	printf("%s: %-4s  %-7s %-12s ; %s\n", $codes[$i]{adr}, $codes[$i]{hex}, $marks{$codes[$i]{adr}}, $codes[$i]{asm}, $codes[$i]{comment});
}



#######################################
# UP ein Befehl reassemblieren
#######################################

sub reass {
	my %code;
	$code{adr} = sprintf ("%.2X", $pc);
	$code{hex} = $hexcode[$pc];
	$code{flagsyntax} = '';
	$code{flag1} = ''; 
	$code{flag1s} = '';
	
	# print "-------------$pc: $hexcode[$pc]\n";
	
	if    ($code{hex} eq '0') { $code{asm} = 'KA'; $code{syntax}= 'A := Key'; $code{comment}= 'Tastendruck nach A bernehmen'; $code{flag1} = 'wenn kein Tastendruck'; $code{flag1s} = 'IF not keypressed THEN '; }
	elsif ($code{hex} eq '1') { $code{asm} = 'AO'; $code{syntax}= 'O := A'; $code{comment}= 'A an Siebensegmentanzeige ausgeben'; }
	elsif ($code{hex} eq '2') { $code{asm} = 'CH'; $code{syntax}= 'A <=> B;  Y <=> Z'; $code{comment}= 'Tausche A und B sowie Y und Z'; }
	elsif ($code{hex} eq '3') { $code{asm} = 'CY'; $code{syntax}= 'A <=> Y'; $code{comment}= 'Austausch der Inhalte von A und Y'; }
	elsif ($code{hex} eq '4') { $code{asm} = 'AM'; $code{syntax}= 'mem[Y] := A'; $code{comment}= 'A in RAM-Adr M[Y] kopieren'; }
	elsif ($code{hex} eq '5') { $code{asm} = 'MA'; $code{syntax}= 'A := mem[Y]'; $code{comment}= 'RAM-Adr M[Y] in A zurcklesen'; }
	elsif ($code{hex} eq '6') { $code{asm} = 'M+'; $code{syntax}= 'A := mem[Y]+A'; $code{comment}= 'RAM-Adr M[Y] zu A addieren'; $code{flag1} = 'bei berlauf'; $code{flag1s} = 'IF overflow THEN '; }
	elsif ($code{hex} eq '7') { $code{asm} = 'M-'; $code{syntax}= 'A := mem[Y]-A'; $code{comment}= 'RAM-Adr M[Y] von A subtrahieren'; $code{flag1} = 'bei berlauf'; $code{flag1s} = 'IF overflow THEN '; }
	elsif ($code{hex} eq '8') { $pc++; $i = $hexcode[$pc]; $code{hex} .= " $i";
								$code{asm} = "TIA $i"; $i = hex $i; $code{syntax}= "A := $i"; $code{comment}= "Konstante $i in A laden"; $code{flag1} = ''; $code{flag1s} = ''; }
	elsif ($code{hex} eq '9') { $pc++; $i = $hexcode[$pc]; $code{hex} .= " $i";
								$code{asm} = "AIA $i"; $i = hex $i; $code{syntax}= "A := A+$i"; $code{comment}= "$i zu A addieren"; $code{flag1} = 'bei berlauf'; $code{flag1s} = 'IF overflow THEN '; }
	elsif ($code{hex} eq 'A') { $pc++; $i = $hexcode[$pc]; $code{hex} .= " $i";
								$code{asm} = "TIY $i"; $i = hex $i; $code{syntax}= "Y := $i"; $code{comment}= "Konstante $i in Y laden"; }
	elsif ($code{hex} eq 'B') { $pc++; $i = $hexcode[$pc]; $code{hex} .= " $i";
								$code{asm} = "AIY $i"; $i = hex $i; $code{syntax}= "Y := Y+$i"; $code{comment}= "$i zu Y addieren"; $code{flag1} = 'bei berlauf'; $code{flag1s} = 'IF overflow THEN '; }
	elsif ($code{hex} eq 'C') { $pc++; $i = $hexcode[$pc]; $code{hex} .= " $i";
								$code{asm} = "CIA $i"; $i = hex $i; $code{syntax}= "A==$i ?"; $code{comment}= "A mit $i vergleichen"; $code{flag1} = 'wenn nicht gleich'; $code{flag1s} = 'IF not equal THEN '; }
	elsif ($code{hex} eq 'D') { $pc++; $i = $hexcode[$pc]; $code{hex} .= " $i";
								$code{asm} = "CIY $i"; $i = hex $i; $code{syntax}= "Y==$i ?"; $code{comment}= "Y mit $i vergleichen"; $code{flag1} = 'wenn nicht gleich'; $code{flag1s} = 'IF not equal THEN '; }
	elsif ($code{hex} eq 'E') { $pc++; $i = $hexcode[$pc]; $code{hex} .= " $i";
								# $code{asm} = "CAL $i"; $code{syntax}= "CAL SUB_$i"; $code{comment}= 'Unterprogrammaufruf'; $code{flag1} = ''; 
								if ($i eq '0')    { $code{syntax}= "CAL RSTO"; $code{comment}= 'Siebensegmentanzeige lschen'; }
								elsif ($i eq '1') { $code{syntax}= "CAL SETR(Y)"; $code{comment}= 'Einzelne LED einschalten. In Y wird die Nummer der LED (0-6) bergeben'; }
								elsif ($i eq '2') { $code{syntax}= "CAL RSTR(Y)"; $code{comment}= 'Einzelne LED ausschalten. In Y wird die Nummer der LED (0-6) bergeben'; }
								elsif ($i eq '4') { $code{syntax}= "CAL CMPL(A)"; $code{comment}= 'Komplement des A-Registers (aus F wird 0)'; }
								elsif ($i eq '5') { $code{syntax}= "CAL CHNG"; $code{comment}= "Inhalte der Register A,B,Y,Z mit A',B',Y',Z' tauschen"; }
								elsif ($i eq '6') { $code{syntax}= "CAL SIFT(A)"; $code{comment}= 'A-Register bitweise nach rechts schieben'; $code{flag1} = 'wenn das rechte Bit 0 war'; $code{flag1s} = 'IF rightbit==0 THEN ';}
								elsif ($i eq '7') { $code{syntax}= "CAL ENDS"; $code{comment}= 'Ende-Sound'; }
								elsif ($i eq '8') { $code{syntax}= "CAL ERRS"; $code{comment}= 'Error-Sound'; }
								elsif ($i eq '9') { $code{syntax}= "CAL SHTS"; $code{comment}= 'Kurzer Ton'; }
								elsif ($i eq 'A') { $code{syntax}= "CAL LONS"; $code{comment}= 'Langer Ton'; }
								elsif ($i eq 'B') { $code{syntax}= "CAL SUND(A)"; $code{comment}= 'Note spielen, die im A-Register bergeben wird (1..E)'; }
								elsif ($i eq 'C') { $code{syntax}= "CAL TIMR(A)"; $code{comment}= '(A + 1) * 0,1 Sekunden warten'; }
								elsif ($i eq 'D') { $code{syntax}= "CAL DSPR"; $code{comment}= 'Ausgabe der RAM-Adressen M[F] (high) und M[E] (low) an die LEDs'; }
								elsif ($i eq 'E') { $code{syntax}= "CAL DEM-"; $code{comment}= ''; }
								elsif ($i eq 'F') { $code{syntax}= "CAL DEM+"; $code{comment}= ''; }
								else {$code{syntax} = "??? CAL $i"}
								$code{asm} = $code{syntax};
								}
	elsif ($code{hex} eq 'F') { $pc++; $i = $hexcode[$pc]; $pc++; $i .= $hexcode[$pc]; $code{hex} .= " $i";
								$marks{$i} = "LBL_$i"; $syntaxmarks{$i} ||= "label".++$label;
								$code{asm} = "JUMP $marks{$i}"; $code{syntax}= "GOTO $syntaxmarks{$i}"; 
								#$code{comment}= "Sprung zur Adresse $marks{$i}"; 
								# Bedingung vom vorherigen Befehl anhngen
								$code{flagsyntax} = $codes[$#codes]{flag1s};
								#$code{comment} .= ' '.$codes[$#codes]{flag1} if $codes[$#codes]{flag1}  
								$code{comment} = $codes[$#codes]{flag1};
								$code{comment} .= ' ' if $code{comment};
								$code{comment} .= "Sprung zur Adresse $marks{$i}"; 
								$marks{$i} .= ':'; 
							}
	# Fehlerfall
	else { $code{asm} = "??? $hexcode[$pc]" }
								  
	# print Dumper(\%code);

	push @codes, \%code;
	$pc++;
}

